/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package net.jazz.rtcext.workitem.extensions.ide.ui;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.jazz.rtcext.workitem.extensions.common.IBuildOnStateChangeDefinitions;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.fieldassist.ControlDecoration;
import org.eclipse.jface.fieldassist.FieldDecorationRegistry;
import org.eclipse.jface.layout.GridLayoutFactory;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ComboViewer;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.forms.widgets.ExpandableComposite;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.Section;

import com.ibm.team.build.internal.client.ITeamBuildRecordClient;
import com.ibm.team.build.internal.common.model.dto.IBuildDefinitionStatusRecord;
import com.ibm.team.foundation.common.util.IMemento;
import com.ibm.team.process.common.IProcessArea;
import com.ibm.team.process.common.IProcessDefinition;
import com.ibm.team.process.common.IProcessItem;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.process.ide.ui.OperationDetailsAspectEditor;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.internal.workflow.WorkflowManager;
import com.ibm.team.workitem.common.model.IState;
import com.ibm.team.workitem.common.model.IWorkItemType;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

/**
 * The configuration information for an operation participant is stored in the
 * project or team area's process configuration XML. The process framework
 * manages the overall document. For extensions from other components, like this
 * one, the process framework delegates editing of the relevant XML, an aspect,
 * to an aspect editor. The process framework is able to learn from our schema
 * exactly which aspect of the XML to delegate to this editor.
 * 
 * This class is an aspect editor for the details of the build on state change
 * follow-up action for work item save. The user can select ids from comboboxes.
 */
public class BuildOnStateChangeAspectEditor extends
		OperationDetailsAspectEditor {
	/*
	 * These constants are used to properly space the controls. The numbers are
	 * pixels that will be placed between the cells of all grid layouts.
	 */
	private static final int HORIZONTAL_SPACING = 10;
	private static final int VERTICAL_SPACING = 8;

	/*
	 * Used to read and write the XML aspect. It also has get and set methods
	 * for the ids. These get and set methods are used by this aspect editor.
	 * The aspect editor does not have any XML considerations. That is handled
	 * by the model and the process framework used by the model.
	 */
	private BuildOnStateChangeModel fModel;

	/*
	 * These fields hold a subset of the controls created for this aspect
	 * editor. Only fields that the rest of the code has to interact with are
	 * stored. Other controls, such as canvases are created but not stored. The
	 * comboboxes are for selection only. The user does not need to type into
	 * them. The decorations are small icons shown next to a combobox when there
	 * is a validation problem. The message describing the validation problem is
	 * available by hovering over the decoration.
	 */
	private ComboViewer fWorkItemTypeIdField;
	private ControlDecoration fWorkItemTypeIdDecoration;
	private ComboViewer fWorkItemStateIdField;
	private ControlDecoration fWorkItemStateIdDecoration;
	private ComboViewer fBuildDefIdField;
	private ControlDecoration fBuildDefIdDecoration;

	/*
	 * These arrays hold the choices for the comboboxes. Note that that each
	 * type contains an array of states for the states combobox. The work item
	 * types (including the states for each type) and build definitions are
	 * retrieved from the repository. Only the values of interest to this aspect
	 * editor are retained (see the type definitions used here and defined later
	 * in this file).
	 */
	private WorkItemType[] fWorkItemTypes;
	private BuildDef[] fBuildDefinitions;

	/*
	 * These are services that this aspect editor needs repeatedly. This aspect
	 * editor will use these services more than once to get all the work item
	 * type and state information that is required, so they are cached here. The
	 * aspect editor retrieves all the build definitions just once so there is
	 * no need to cache the service.
	 */
	private IWorkItemCommon fWorkItemCommon;
	private WorkflowManager fWorkflowManager;

	/**
	 * Called by the process editor framework when the user decides to edit the
	 * settings for the build on state change operation participant.
	 * 
	 * @param parent
	 *            the composite provided by the framework to hold the controls.
	 *            This method must set the appropriate layout on this composite.
	 * @param toolkit
	 *            a control factory provided by the framework. The process
	 *            framework specializes the Eclipse UI form toolkit so that the
	 *            underlying controls behave properly in the process
	 *            configuration editor. All controls are either created directly
	 *            from the toolkit or passed to the toolkit's adapt method right
	 *            after creation. This makes the aspect editor creator's job
	 *            much easier with regard to the proper process configuration
	 *            editor look and feel. Note that the control decorations are
	 *            not adapted to the toolkit.
	 */
	public void createControl(final Composite parent, FormToolkit toolkit) {
		/*
		 * Create the controls for our aspect editor. There are the controls to
		 * select the work item type and state, followed by one to select the
		 * build definition id.
		 */
		createTriggerControls(parent, toolkit);
		createBuildControls(parent, toolkit);

		/*
		 * The grid data added to controls by the form toolkit along with the
		 * JFace generated grid layout works well in this case and is very easy
		 * to use. Just call generateLayout after all the controls are created.
		 * The generated layout is set on the parent composite passed to this
		 * method, and thus satisfies the constraint from the process framework.
		 * When creating an aspect editor, this technique for layout should be
		 * tried first. Only if it does not work properly should you try
		 * creating your own grid data for your controls. And even then, you can
		 * simply add grid data to one or a small number of controls and let the
		 * form toolkit and JFace layout generation handle the rest.
		 */
		GridLayoutFactory layoutFactory = GridLayoutFactory.fillDefaults()
				.spacing(HORIZONTAL_SPACING, VERTICAL_SPACING).numColumns(1);
		layoutFactory.generateLayout(parent);

		/*
		 * Initialize the user interface controls from the build on state change
		 * model. This also loads the combobox lists.
		 */
		initUI();
	}

	/**
	 * The trigger section consists primarily of two comboboxes. One for the
	 * work item type and one for the work item state. Both are required and
	 * both are read only, that is, selection from the list is the only way to
	 * set the values. The list values are obtained from the project area.
	 * 
	 * @param parent
	 *            passed through from createControl. See description there.
	 * @param toolkit
	 *            passed through from createControl. See description there.
	 */
	private void createTriggerControls(Composite parent, FormToolkit toolkit) {
		/*
		 * The section to hold our trigger controls has a simple title bar and a
		 * composite as its client area.
		 */
		Section triggerSection = toolkit.createSection(parent,
				ExpandableComposite.TITLE_BAR);
		Composite triggerBody = toolkit.createComposite(triggerSection);
		triggerSection.setClient(triggerBody);
		triggerSection.setText("Work Item Trigger");

		/*
		 * The work item type id selection controls consist of a label and a
		 * combobox. The combobox is decorated with a small icon to show
		 * validation errors and other conditions.
		 */
		toolkit.createLabel(triggerBody, "Type Id:");
		fWorkItemTypeIdField = new ComboViewer(triggerBody);
		toolkit.adapt(fWorkItemTypeIdField.getCombo(), true, false);
		fWorkItemTypeIdDecoration = new ControlDecoration(fWorkItemTypeIdField
				.getCombo(), SWT.TOP | SWT.LEFT);

		/*
		 * The input for the combobox list will be set to an array of
		 * WorkItemType (see definition below). The array content provider adds
		 * one entry to the combobox list for each input array element and calls
		 * toString on each element to get the display value. The WorkItemType
		 * class specializes the toString method for this purpose.
		 */
		fWorkItemTypeIdField.setContentProvider(new ArrayContentProvider());
		fWorkItemTypeIdField.setLabelProvider(new LabelProvider());

		/*
		 * When the work item type selection changes, change to the appropriate
		 * set of states on the state selection combobox and validate all
		 * selections.
		 */
		fWorkItemTypeIdField
				.addSelectionChangedListener(new ISelectionChangedListener() {
					public void selectionChanged(SelectionChangedEvent e) {
						// A ComboViewer always uses a structured selection.
						IStructuredSelection sel = (IStructuredSelection) e
								.getSelection();
						WorkItemType newType = (WorkItemType) sel
								.getFirstElement();
						// If the type really is a change (will not be during
						// initialization of the UI from the model)...
						if (getModel()
								.setWorkItemTypeId(
										(newType == null) ? IBuildOnStateChangeDefinitions.EMPTY_ID
												: newType.getId())) {
							// Change to the new set of states.
							initStates(newType);
							// Editor is now dirty.
							setDirty();
							// Validate all the selections.
							validateSelections();
						}
					}
				});

		/*
		 * The work item state id selection controls consist of a label and a
		 * combobox. The combobox is decorated with a small icon to show
		 * validation errors and other conditions.
		 */
		toolkit.createLabel(triggerBody, "State Id:");
		fWorkItemStateIdField = new ComboViewer(triggerBody);
		toolkit.adapt(fWorkItemStateIdField.getCombo(), true, false);
		fWorkItemStateIdDecoration = new ControlDecoration(
				fWorkItemStateIdField.getCombo(), SWT.TOP | SWT.LEFT);
		fWorkItemStateIdField.setContentProvider(new ArrayContentProvider());
		fWorkItemStateIdField.setLabelProvider(new LabelProvider());

		/*
		 * When the work item state selection changes validate all selections.
		 */
		fWorkItemStateIdField
				.addSelectionChangedListener(new ISelectionChangedListener() {
					public void selectionChanged(SelectionChangedEvent e) {
						// A ComboViewer always uses a structured selection.
						IStructuredSelection sel = (IStructuredSelection) e
								.getSelection();
						WorkItemState wiState = (WorkItemState) sel
								.getFirstElement();
						// If the state really is a change (will not be during
						// initialization of the UI from the model)...
						if (getModel()
								.setWorkItemStateId(
										(wiState == null) ? IBuildOnStateChangeDefinitions.EMPTY_ID
												: wiState.getId())) {
							// Editor is now dirty.
							setDirty();
							// Validate all the selections.
							validateSelections();
						}
					}
				});

		/*
		 * As with the parent composite, the grid data from the toolkit and the
		 * generated layout from JFace are fine. Note that the nested composite
		 * created here has its layout generated before the parent layout is
		 * generated. This is as recommended by JFace.
		 */
		GridLayoutFactory layoutFactory = GridLayoutFactory.fillDefaults()
				.spacing(HORIZONTAL_SPACING, VERTICAL_SPACING).numColumns(2);
		layoutFactory.generateLayout(triggerBody);
	}

	/**
	 * The build section consists primarily of one combobox for the build
	 * definition id. It is required and read only, that is, selection from the
	 * list is the only way to set the value. The list values are obtained from
	 * the project area.
	 * 
	 * @param parent
	 *            passed through from createControl. See description there.
	 * @param toolkit
	 *            passed through from createControl. See description there.
	 */
	private void createBuildControls(Composite parent, FormToolkit toolkit) {
		/*
		 * The section to hold our build controls has a simple title bar and a
		 * composite as its client area.
		 */
		Section buildSection = toolkit.createSection(parent,
				ExpandableComposite.TITLE_BAR);
		Composite buildBody = toolkit.createComposite(buildSection);
		buildSection.setClient(buildBody);
		buildSection.setText("Build Definition");

		/*
		 * The build definition id selection controls consist of a label and a
		 * combobox. The combobox is decorated with a small icon to show
		 * validation errors and other conditions.
		 */
		toolkit.createLabel(buildBody, "Id:");
		fBuildDefIdField = new ComboViewer(buildBody);
		toolkit.adapt(fBuildDefIdField.getCombo(), true, false);
		fBuildDefIdDecoration = new ControlDecoration(fBuildDefIdField
				.getCombo(), SWT.TOP | SWT.LEFT);
		fBuildDefIdField.setContentProvider(new ArrayContentProvider());
		fBuildDefIdField.setLabelProvider(new LabelProvider());

		/*
		 * When the build definition selection changes validate all selections.
		 */
		fBuildDefIdField
				.addSelectionChangedListener(new ISelectionChangedListener() {
					public void selectionChanged(SelectionChangedEvent e) {
						// A ComboViewer always uses a structured selection.
						IStructuredSelection sel = (IStructuredSelection) e
								.getSelection();
						BuildDef buildDef = (BuildDef) sel.getFirstElement();
						// If the build really is a change (will not be during
						// initialization of the UI from the model)...
						if (getModel()
								.setBuildDefId(
										(buildDef == null) ? IBuildOnStateChangeDefinitions.EMPTY_ID
												: buildDef.getId())) {
							// Editor is now dirty.
							setDirty();
							// Validate all the selections.
							validateSelections();
						}
					}
				});

		/*
		 * As with the parent composite, the grid data from the toolkit and the
		 * generated layout from JFace are fine. Note that the nested composite
		 * created here has its layout generated before the parent layout is
		 * generated. This is as recommended by JFace.
		 */
		GridLayoutFactory layoutFactory = GridLayoutFactory.fillDefaults()
				.spacing(HORIZONTAL_SPACING, VERTICAL_SPACING).numColumns(2);
		layoutFactory.generateLayout(buildBody);
	}

	/**
	 * Initialize the UI from the build on state change model. This also loads
	 * the comboxbox lists.
	 */
	private void initUI() {
		/*
		 * Get the work item types for the project area and load them into the
		 * combobox list.
		 */
		WorkItemType[] wiTypes = getWorkItemTypes();
		fWorkItemTypeIdField.setInput(wiTypes);

		/*
		 * Get the current work item type id from the model and search through
		 * the work item types just obtained from the project area. If a match
		 * is found, set it as the current selection.
		 */
		String modelId = getModel().getWorkItemTypeId();
		WorkItemType selectedWIType = null;
		for (int i = 0; i < (wiTypes.length) && (selectedWIType == null); i++) {
			if (modelId.equals(wiTypes[i].getId())) {
				selectedWIType = wiTypes[i];
				fWorkItemTypeIdField.setSelection(new StructuredSelection(
						selectedWIType));
			}
		}

		/*
		 * Initializing the states UI is in a separate method because it also
		 * needs to be done whenever the work item type selection is changed.
		 * See the initStates method for details.
		 */
		initStates(selectedWIType);

		/*
		 * Get the build definitions for the project area and load them into the
		 * combobox list.
		 */
		BuildDef[] buildDefs = getBuildDefinitions();
		fBuildDefIdField.setInput(buildDefs);

		/*
		 * Get the current build definition id from the model and search through
		 * the build definitions just obtained from the project area. If a match
		 * is found, set it as the current selection.
		 */
		modelId = getModel().getBuildDefId();
		BuildDef selectedBuildDef = null;
		for (int i = 0; (i < buildDefs.length) && (selectedBuildDef == null); i++) {
			if (modelId.equals(buildDefs[i].getId())) {
				selectedBuildDef = buildDefs[i];
				fBuildDefIdField.setSelection(new StructuredSelection(
						selectedBuildDef));
			}
		}

		/*
		 * Finally, validate all the combobox selections. See the
		 * validateSelections method for details.
		 */
		validateSelections();
	}

	/**
	 * This method is called to initialize the work item states UI when the
	 * aspect editor is first created (from initUI) and whenever the work item
	 * type changes (from the selection changed listener attached to the work
	 * item type combobox). It checks to see if the current state is valid for
	 * the new type and adjusts the model and current work item state selection
	 * if needed to keep the model and UI consistent.
	 * 
	 * @param selectedWIType
	 *            the currently selected work item type. It is used to get the
	 *            valid set of states.
	 */
	private void initStates(WorkItemType selectedWIType) {
		/*
		 * If there is a currently selected work item type...
		 */
		if (selectedWIType != null) {
			/*
			 * Get the current work item state selection. It will be checked
			 * against the new list of valid states for the currently selected
			 * work item type. Need to get it before setting the new list of
			 * states on the combobox, since setting the new list could change
			 * the current selection.
			 */
			WorkItemState currentState = (WorkItemState) ((IStructuredSelection) fWorkItemStateIdField
					.getSelection()).getFirstElement();

			/*
			 * Get the new list of valid states from the currently selected work
			 * item type and set it as the new list for the combobox.
			 */
			WorkItemState[] wiStates = selectedWIType.getStates();
			fWorkItemStateIdField.setInput(wiStates);

			/*
			 * If a prior state was obtained above, that is, there was a current
			 * state selection when this method was called, search the new list
			 * of valid states for a match. Note that the search is by state
			 * name and not state id. This is intentional. Users generally think
			 * in terms of the names, but our code needs the id in the model for
			 * the participant to work. Also, it is common for two workflows to
			 * have states with the same name but different ids. For example in
			 * the Scrum template, both Epic and Story have a Done state, but
			 * have different ids for that state. We want to follow the user's
			 * mental model here.
			 */
			if (currentState != null) {
				String currentName = currentState.getName();
				WorkItemState selectedState = null;
				for (int i = 0; (i < wiStates.length)
						&& (selectedState == null); i++) {
					if (currentName.equals(wiStates[i].getName())) {
						/*
						 * There is a match and can stop searching. Setting the
						 * new state as the current selection on the combobox
						 * will have the side affect of updating the model if
						 * needed via the selection changed listener on the
						 * combobox. An update will happen if the two states
						 * (the old and new one with the same name) have
						 * different ids. If they have the same id, no update of
						 * the model will happen.
						 */
						selectedState = wiStates[i];
						fWorkItemStateIdField
								.setSelection(new StructuredSelection(
										selectedState));
					}
				}
			} else {
				/*
				 * There was no prior state. In this case, get the work item
				 * state id from the model and search through the new set of
				 * work item state ids. If a match is found, set it as the
				 * current selection.
				 */
				String modelId = getModel().getWorkItemStateId();
				WorkItemState selectedState = null;
				for (int i = 0; (i < wiStates.length)
						&& (selectedState == null); i++) {
					if (modelId.equals(wiStates[i].getId())) {
						selectedState = wiStates[i];
						fWorkItemStateIdField
								.setSelection(new StructuredSelection(
										selectedState));
					}
				}
			}
		} else {
			/*
			 * There is no current work item type selected, so there is no valid
			 * list of work item states.
			 */
			fWorkItemStateIdField.setInput(null);
		}
	}

	/**
	 * Validate the selections for all three ids. By far, the most common case
	 * is that this participant will be configured in a project or team area. In
	 * these cases, it is straight forward to obtain the valid ids for
	 * selection. The rare case is configuring the participant in a process
	 * template. In this case, it is not possible to know the valid build
	 * definitions and getting the others is not straight forward. In this case
	 * the user is directed to the XML.
	 * 
	 * Note that a ComboViewer always uses a structured selection and that the
	 * initialization and listener code is keeping the model and the combobox
	 * selection in sync. Therefore, the casts are safe and just checking for
	 * selection validates the data.
	 */
	protected void validateSelections() {
		/*
		 * If in a project or team area, validate the selections.
		 */
		if (getAspect().getProcessContainerWorkingCopy()
				.getUnderlyingProcessItem() instanceof IProcessArea) {
			/*
			 * The work item type id must be one from the project area.
			 */
			boolean noWITypeSelection = ((IStructuredSelection) fWorkItemTypeIdField
					.getSelection()).getFirstElement() == null;
			if (noWITypeSelection) {
				setValidationMessage(true, fWorkItemTypeIdDecoration,
						"The work item type is not valid. Select one from the list.");
			} else {
				setValidationMessage(true, fWorkItemTypeIdDecoration, null);
			}

			/*
			 * The state id must be valid for the selected work item type.
			 */
			if (((IStructuredSelection) fWorkItemStateIdField.getSelection())
					.getFirstElement() == null) {
				if (noWITypeSelection)
					setValidationMessage(true, fWorkItemStateIdDecoration,
							"Select a work item type and then a state.");
				else
					setValidationMessage(true, fWorkItemStateIdDecoration,
							"The work item state is not valid. Select one from the list.");
			} else {
				setValidationMessage(true, fWorkItemStateIdDecoration, null);
			}

			/*
			 * The build definition id must be one from the project area.
			 */
			if (((IStructuredSelection) fBuildDefIdField.getSelection())
					.getFirstElement() == null) {
				setValidationMessage(true, fBuildDefIdDecoration,
						"The build definition is not valid. Select one from the list.");
			} else {
				setValidationMessage(true, fBuildDefIdDecoration, null);
			}
		} else {
			/*
			 * For a process template, warn that edits must be via the XML.
			 */
			fWorkItemTypeIdDecoration.setImage(FieldDecorationRegistry
					.getDefault().getFieldDecoration(
							FieldDecorationRegistry.DEC_WARNING).getImage());
			fWorkItemTypeIdDecoration
					.setDescriptionText("Search for buildOnStateChange in the process XML to edit all values.");
		}
	}

	/**
	 * Helper to set the proper message and field decoration. Note that this
	 * method assumes any message passed in is an error.
	 * 
	 * @param required
	 *            true if the validated field is required and will be decorated
	 *            with the required icon and message if no error message is
	 *            passed in.
	 * @param decoration
	 *            the decoration control to set
	 * @param msg
	 *            the message to set. Can be null to clear a message.
	 */
	protected void setValidationMessage(boolean required,
			ControlDecoration decoration, String msg) {
		/*
		 * If there is a message to set update the decorations icon to the error
		 * icon and set the message for hover.
		 */
		if (msg != null) {
			decoration.setImage(FieldDecorationRegistry.getDefault()
					.getFieldDecoration(FieldDecorationRegistry.DEC_ERROR)
					.getImage());
			decoration.setDescriptionText(msg);
		} else {
			/*
			 * There is no message, so clear the decoration. Clearing means
			 * setting the required icon and message on a required field.
			 */
			if (required) {
				decoration.setImage(FieldDecorationRegistry.getDefault()
						.getFieldDecoration(
								FieldDecorationRegistry.DEC_REQUIRED)
						.getImage());
				decoration.setDescriptionText("Required");
			} else {
				decoration.hide();
				decoration.setImage(null);
				decoration.show();
			}
		}
	}

	/**
	 * Access method for the build on state change model.
	 * 
	 * @return the model
	 */
	protected BuildOnStateChangeModel getModel() {
		if (fModel == null)
			fModel = new BuildOnStateChangeModel();
		return fModel;
	}

	/**
	 * Called by the process framework before createControl. The aspect editor
	 * is to read the model from the memento.
	 * 
	 * @param memento
	 *            the root memento for the participant's XML aspect
	 */
	public void restoreState(IMemento memento) {
		getModel().readFrom(memento);
	}

	/**
	 * Called by the process framework to save the state to the XML process
	 * configuration.
	 * 
	 * @param memento
	 *            the root memento for the participant's XML aspect
	 */
	public boolean saveState(IMemento memento) {
		return getModel().saveTo(memento);
	}

	/**
	 * Access method for the work item common service. This service is used to
	 * obtain the work item types from the project area.
	 * 
	 * @return the work item common service
	 */
	protected IWorkItemCommon getWorkItemCommon() {
		if (fWorkItemCommon == null) {
			ITeamRepository teamRepository = (ITeamRepository) getAspect()
					.getProcessContainerWorkingCopy()
					.getUnderlyingProcessItem().getOrigin();
			fWorkItemCommon = (IWorkItemCommon) teamRepository
					.getClientLibrary(IWorkItemCommon.class);
		}
		return fWorkItemCommon;
	}

	/**
	 * Access method for the work item workflow manager. This service is used to
	 * obtain the valid states for each work item type.
	 * 
	 * @return the workflow manager
	 */
	protected WorkflowManager getWorkflowManager() {
		if (fWorkflowManager == null) {
			fWorkflowManager = (WorkflowManager) getWorkItemCommon()
					.getWorkflowManager();
		}
		return fWorkflowManager;
	}

	/**
	 * This editor does not need to dispose its controls the container takes
	 * care of that. The process framework calls this method so that an aspect
	 * editor can free any scarce resources that it may be holding. This editor
	 * does not have any.
	 */
	public void dispose() {
	}

	/**
	 * Gets the work item types from the project area and creates our
	 * WorkItemType elements. These elements will be set as the input for the
	 * work item type combobox. This method uses the getStatesForTypeCategory
	 * method to obtain the valid states for each work item type.
	 * 
	 * @return the work item types array
	 */
	protected WorkItemType[] getWorkItemTypes() {
		if (fWorkItemTypes == null) {
			/*
			 * The underlying process item (project area, team area or process
			 * template) is checked for its type. If it is a process area
			 * (project area or team area), the project area itself or the
			 * containing project area (in the case of a team area) is used to
			 * find the work item types.
			 */
			final IProcessItem processItem = getAspect()
					.getProcessContainerWorkingCopy()
					.getUnderlyingProcessItem();
			if (processItem instanceof IProcessArea) {
				/*
				 * We are editing a project or team area.
				 */
				try {
					/*
					 * This could hit the server so move off UI thread.
					 */
					PlatformUI.getWorkbench().getActiveWorkbenchWindow().run(
							true, true, new IRunnableWithProgress() {
								public void run(IProgressMonitor monitor)
										throws InvocationTargetException,
										InterruptedException {
									try {
										/*
										 * Obtain the project area that is being
										 * configured and then get its
										 * configured work item types.
										 */
										IProjectAreaHandle projectArea = ((IProcessArea) processItem)
												.getProjectArea();
										List<IWorkItemType> rawWorkitemTypes = getWorkItemCommon()
												.findWorkItemTypes(projectArea,
														monitor);

										/*
										 * The result array is the same sized to
										 * match the number of work item types
										 * in the project area.
										 */
										fWorkItemTypes = new WorkItemType[rawWorkitemTypes
												.size()];

										/*
										 * Each work item type belongs to work
										 * item type category and all work item
										 * types in the same category use the
										 * same workflow states. This map is
										 * used to avoid accessing the same set
										 * of states repeatedly. The same array
										 * of WorkItemStates can be shared by
										 * several instances of WorkItemType.
										 */
										Map<String, WorkItemState[]> wftCatToStatesMap = new HashMap<String, WorkItemState[]>();

										/*
										 * For each work item type...
										 */
										int i = 0;
										for (IWorkItemType wit : rawWorkitemTypes) {
											/*
											 * Get the type's category and check
											 * for an existing set of states for
											 * the type.
											 */
											String witCategoryId = wit
													.getCategory();
											WorkItemState[] states = wftCatToStatesMap
													.get(witCategoryId);
											/*
											 * If no cached states, get a new
											 * set and cache it.
											 */
											if (states == null) {
												states = getStatesForTypeCategory(
														projectArea,
														witCategoryId, monitor);
												wftCatToStatesMap.put(
														witCategoryId, states);
											}

											/*
											 * Create the WorkItemType instance
											 * from the id, name and states.
											 */
											fWorkItemTypes[i] = new WorkItemType(
													wit.getIdentifier(), wit
															.getDisplayName(),
													states);
											i++;
										}
									} catch (TeamRepositoryException e) {
										WorkitemExtensionsPlugin
												.getDefault()
												.log(
														"Error retrieving workitem types.",
														e);
									}
								}
							});
				} catch (InvocationTargetException e) {
					WorkitemExtensionsPlugin.getDefault().log(
							"Error retrieving workitem types.", e);
				} catch (InterruptedException e) {
					WorkitemExtensionsPlugin.getDefault().log(
							"Error retrieving workitem types.", e);
				}
			} else if (processItem instanceof IProcessDefinition) {
				/*
				 * We are editing a process template. The default empty array is
				 * fine. Handled below along with any post exception processing.
				 */
			} else {
				throw new UnsupportedOperationException(
						"Process item is neither process area nor process definition.");
			}

			/*
			 * In the case of a process template or a caught exception, the work
			 * items types array will still be null. For the process definition
			 * case, the user will be directed to the XML. For the exception
			 * cases, the error has been logged. In any event, a default empty
			 * array will work.
			 */
			if (fWorkItemTypes == null)
				fWorkItemTypes = new WorkItemType[0];
		}

		return fWorkItemTypes;
	}

	/**
	 * Used by the getWorkItemTypes method to obtain an array of work item
	 * states for a given work item type category.
	 * 
	 * @param projectArea
	 *            the project area being configured
	 * @param witCategoryId
	 *            the id of the work item type category for which the state
	 *            array is needed
	 * @param monitor
	 *            a progress monitor that is passed to the underlying service
	 *            methods
	 * @return the work item states for the category
	 */
	protected WorkItemState[] getStatesForTypeCategory(
			IProjectAreaHandle projectArea, String witCategoryId,
			IProgressMonitor monitor) {
		/*
		 * Get the workflow id for the supplied category from the workflow
		 * manager service and then get the detailed workflow information.
		 */
		WorkflowManager wfm = getWorkflowManager();
		String wfId = wfm.getWorkflowIdForCategory(projectArea, witCategoryId,
				false, monitor);
		IWorkflowInfo wfInfo = wfm.getWorkflowInfo(projectArea, wfId, false,
				monitor);

		/*
		 * Obtain all the state ids in the workflow and create our result array
		 * of the same length.
		 */
		Identifier<IState>[] rawStates = wfInfo.getAllStateIds();
		WorkItemState[] states = new WorkItemState[rawStates.length];
		/*
		 * For each state obtained for the workflow, create a WorkItemState
		 * instance from the id and name.
		 */
		for (int i = 0; i < rawStates.length; i++) {
			states[i] = new WorkItemState(rawStates[i].getStringIdentifier(),
					wfInfo.getStateName(rawStates[i]));
		}

		return states;
	}

	/**
	 * Gets the build definitions for the project area and maps them by id.
	 * There must be an active workbench window, so call this from a selection
	 * event handler or similar location.
	 * 
	 * @return the build definitions sorted by id
	 */
	private BuildDef[] getBuildDefinitions() {
		if (fBuildDefinitions == null) {
			final IProcessItem processItem = getAspect()
					.getProcessContainerWorkingCopy()
					.getUnderlyingProcessItem();
			if (processItem instanceof IProcessArea) {
				/*
				 * We are editing a project or team area.
				 */
				try {
					// This could hit the server so move off UI thread.
					PlatformUI.getWorkbench().getActiveWorkbenchWindow().run(
							true, true, new IRunnableWithProgress() {
								public void run(IProgressMonitor monitor)
										throws InvocationTargetException,
										InterruptedException {
									IProcessArea processArea = (IProcessArea) processItem;
									ITeamRepository teamRepository = (ITeamRepository) processItem
											.getOrigin();
									ITeamBuildRecordClient buildClient = (ITeamBuildRecordClient) teamRepository
											.getClientLibrary(ITeamBuildRecordClient.class);
									try {
										IBuildDefinitionStatusRecord[] buildStatRecords = buildClient
												.getBuildDefinitionStatusRecords(
														new IProjectAreaHandle[] { processArea
																.getProjectArea() },
														monitor);
										fBuildDefinitions = new BuildDef[buildStatRecords.length];
										int i = 0;
										for (IBuildDefinitionStatusRecord bd : buildStatRecords) {
											fBuildDefinitions[i] = new BuildDef(
													bd.getBuildDefinition()
															.getId());
											i++;
										}
									} catch (TeamRepositoryException e) {
										WorkitemExtensionsPlugin
												.getDefault()
												.log(
														"Error retrieving build definitions.",
														e);
									}
								}
							});
				} catch (InvocationTargetException e) {
					WorkitemExtensionsPlugin.getDefault().log(
							"Error retrieving build definitions.", e);
				} catch (InterruptedException e) {
					WorkitemExtensionsPlugin.getDefault().log(
							"Error retrieving build definitions.", e);
				}
			} else if (processItem instanceof IProcessDefinition) {
				/*
				 * We are editing a process template, so there are no build
				 * definitions. Handled below along with any post exception
				 * processing.
				 */
			} else {
				throw new UnsupportedOperationException(
						"Process item is neither process area nor process definition.");
			}

			/*
			 * In the case of a process template or a caught exception, the work
			 * items types array will still be null. For the process definition
			 * case, the user will be directed to the XML. For the exception
			 * cases, the error has been logged. In any event, a default empty
			 * array will work.
			 */
			if (fBuildDefinitions == null)
				fBuildDefinitions = new BuildDef[0];
		}

		return fBuildDefinitions;
	}

	/**
	 * Builds the proper display name for selection dialogs, "name (id)".
	 * 
	 * @param name
	 *            the item's name
	 * @param id
	 *            the item's id
	 * @return the display name
	 */
	protected static String buildDisplayName(String name, String id) {
		return name + " (" + id + ")";
	}

	/**
	 * Represents a work item type for the aspect editor. Most notably, an array
	 * of these is set as the input for the work item type combobox.
	 */
	protected static class WorkItemType {
		/*
		 * The work item type stores the id, name and the state valid for the
		 * work item type's category. The display name is calculated from the id
		 * and name using the buildDisplayName method.
		 */
		private String fId;
		private String fName;
		private WorkItemState[] fStates;
		private String fDisplayName;

		/**
		 * The constuctor takes the id, name and states and stores them. The
		 * display name is also calculated.
		 * 
		 * @param id
		 *            the work item type's id
		 * @param name
		 *            the work item type's name
		 * @param states
		 *            the workflow states for the work item type
		 */
		public WorkItemType(String id, String name, WorkItemState[] states) {
			fId = id;
			fName = name;
			fStates = states;
			fDisplayName = buildDisplayName(name, id);
		}

		/**
		 * Access method for the id.
		 * 
		 * @return the work item type id
		 */
		public String getId() {
			return fId;
		}

		/**
		 * Access method for the name.
		 * 
		 * @return the work item type name
		 */
		public String getName() {
			return fName;
		}

		/**
		 * Access method for the workflow states.
		 * 
		 * @return the workflow states
		 */
		public WorkItemState[] getStates() {
			return fStates;
		}

		/**
		 * The toString method is overridden to work well with the LabelProvider
		 * used on the work item type combobox.
		 * 
		 * @return the display name
		 */
		public String toString() {
			return fDisplayName;
		}
	}

	/**
	 * Represents a work item state for the aspect editor. Most notably, an
	 * array of these is set as the input for the work item state combobox.
	 */
	protected static class WorkItemState {
		/*
		 * The work item state stores the id and name. The display name is
		 * calculated from the id and name using the buildDisplayName method.
		 */
		private String fId;
		private String fName;
		private String fDisplayName;

		/**
		 * The constuctor takes the id and name and stores them. The display
		 * name is also calculated.
		 * 
		 * @param id
		 *            the work item type's id
		 * @param name
		 *            the work item type's name
		 */
		public WorkItemState(String id, String name) {
			fId = id;
			fName = name;
			fDisplayName = buildDisplayName(name, id);
		}

		/**
		 * Access method for the id.
		 * 
		 * @return the work item state id
		 */
		public String getId() {
			return fId;
		}

		/**
		 * Access method for the name.
		 * 
		 * @return the work item state name
		 */
		public String getName() {
			return fName;
		}

		/**
		 * The toString method is overridden to work well with the LabelProvider
		 * used on the work item type combobox.
		 * 
		 * @return the display name
		 */
		public String toString() {
			return fDisplayName;
		}
	}

	/**
	 * Represents a build definition for the aspect editor. Most notably, an
	 * array of these is set as the input for the build definition combobox.
	 */
	private static class BuildDef {
		/*
		 * The build definition stores the id. In this case, the display name is
		 * the id.
		 */
		private String fId;

		/**
		 * The constuctor takes the id stores it.
		 * 
		 * @param id
		 *            the work item type's id
		 */
		public BuildDef(String id) {
			fId = id;
		}

		/**
		 * Access method for the id.
		 * 
		 * @return the build definition id
		 */
		public String getId() {
			return fId;
		}

		/**
		 * The toString method is overridden to work well with the LabelProvider
		 * used on the build definition combobox.
		 * 
		 * @return the display name which in this case is the id
		 */
		public String toString() {
			return getId();
		}
	}

}
